#include "CompileNotValidatedImpl.h"

#include "Components/AutoPtr/AutoPtr.h"
#include "Components/ModuleComponent/ExprComponent.h"
#include "Components/DefaultConfig/DefaultConfig.h"

#include "unicode.h"

int search_operation_precedence(const std::u32string& random) {
	for (auto& s : SLexicalRule.OperationPrecedenceTable) {
		if (s.random == random) {
			return s.precedence;
		}
	}
	return -1;
}

ExprComponent* CompileNotValidatedExpr::performPrefix() {
	auto token = tokenizer->copyOrEnd();
	if (!token.has_value()) {
		pushError(u8"非预期的输入序列终止");
		return nullptr;
	}
	if (token->testRandom(U"::")
		|| token->testRandom(U"++")
		|| token->testRandom(U"--")
		|| token->testRandom(U"*")
		|| token->testRandom(U"&")
		|| token->testRandom(U"+")
		|| token->testRandom(U"-")
		|| token->testRandom(U"!")) {
		AutoPtr<ExprComponent> expr1 = new ExprComponent(ExprComponentKind::unary, *token);
		expr1.data().op = U"prefix." + token->value;
		tokenizer->dropOrEnd();
		auto expr2 = performPrefix();
		if (expr2 == nullptr) {
			return nullptr;
		}
		expr1.data().S[0] = expr2;
		return expr1.take();

	}
	else if (token->isBracket(U"(")
		|| token->isBracket(U"[")) {
		if (token->isBracket(U"(")) {
			return performBracket(U"(", U")");
		}
		else if(token->isBracket(U"[")) {
			return performBracket(U"[", U"]");
		}
		else {
			pushError(u8"程序内部错误");
			return nullptr;
		}
	}
	else {
		return performPostfix(performPrimary());
	}
	return nullptr;
}

ExprComponent* CompileNotValidatedExpr::performPrimary() {
	auto token = tokenizer->copyOrEnd();
	if (!token.has_value()) {
		pushError(u8"非预期的输入序列终止");
		return nullptr;
	}
	if ((token->kind == TokenKind::identifier)
		|| (token->kind == TokenKind::quantity)) {
		if (token->kind == TokenKind::identifier) {
			auto expr = new ExprComponent(ExprComponentKind::identifier, *token);
			expr->identifier = token->value;
			tokenizer->dropOrEnd();
			return expr;
		}
		else if (token->kind == TokenKind::quantity) {
			auto expr = new ExprComponent(ExprComponentKind::quantity, *token);
			expr->quantity = token->value;
			tokenizer->dropOrEnd();
			return expr;
		}
		else {
			pushError(u8"程序内部错误");
			return nullptr;
		}
	}
	else {
		pushError(u8"期望一个标识符");
		return nullptr;
	}
}

ExprComponent* CompileNotValidatedExpr::performPostfix(ExprComponent* expr1) {
	if (expr1 == nullptr) {
		return nullptr;
	}
	auto token = tokenizer->copyOrEnd();
	if (token->testRandom(U"++")
		|| token->testRandom(U"--")) {
		auto expr = new ExprComponent(ExprComponentKind::unary, *token);
		expr->op = U"postfix." + token->value;
		tokenizer->dropOrEnd();
		expr->S[0] = expr1;
		auto expr2 = performPostfix(expr);
		if (expr2 == nullptr) {
			delete expr;
		}
		return expr2;
	}
	else if (token->isBracket(U"(")
		|| token->isBracket(U"[")) {
		auto expr = new ExprComponent(ExprComponentKind::binary, *token);
		expr->S[0] = expr1;
		if (token->isBracket(U"(")) {
			expr->op = U"postfix.()";
			auto expr2 = performBracket(U"(", U")");
			if (expr2 == nullptr) {
				delete expr;
				return nullptr;
			}
			expr->S[1] = expr2;
		}
		else if (token->isBracket(U"[")) {
			expr->op = U"postfix.[]";
			auto expr2 = performBracket(U"[", U"]");
			if (expr2 == nullptr) {
				delete expr;
				return nullptr;
			}
			expr->S[1] = expr2;
			return expr;
		}
		else {
			pushError(u8"程序内部错误");
			return nullptr;
		}
		return performPostfix(expr);
	}
	else if (token->testRandom(U".")
		|| token->testRandom(U"->")) {
		auto expr = new ExprComponent(ExprComponentKind::binary, *token);
		expr->S[0] = expr1;
		expr->op = token->value;
		token = tokenizer->dropAndCopy();
		if (!token.has_value() || !token->isIdentifier()) {
			pushError(u8"不是合法的标识符");
			delete expr;
			return nullptr;
		}
		auto expr2 = new ExprComponent(ExprComponentKind::identifier, *token);
		expr2->identifier = token->value;
		expr->S[1] = expr2;
		token = tokenizer->dropAndCopy();
		return performPostfix(expr);
	}
	else {
		return expr1;
	}
}

ExprComponent* CompileNotValidatedExpr::performBinary(ExprComponent* expr1, int precedence) {
	int p1;
	while (1) {
		auto token = tokenizer->copyOrEnd();
		if (!token.has_value() || (p1 = search_operation_precedence(token->value)) < precedence) {
			return expr1;
		}
		tokenizer->dropOrEnd();
		auto expr_x = performPrefix();
		if (expr_x == nullptr) {
			delete expr1;
			return nullptr;
		}
		auto token2 = tokenizer->copyOrEnd();
		int p2 = search_operation_precedence(token2->value);
		if (p1 < p2) {
			expr_x = performBinary(expr_x, p1 + 1);
			if (expr_x == nullptr) {
				delete expr1;
				return nullptr;
			}
		}
		if (token->testRandom(U"=")) {
			auto expr2 = new ExprComponent(ExprComponentKind::binary, *token);
			expr2->op = token->value;
			expr2->S[0] = expr1;
			auto expr3 = performBinary(expr_x, precedence);
			if (expr3 == nullptr) {
				delete expr2;
				return nullptr;
			}
			expr2->S[1] = expr3;
			expr1 = expr2;
		}
		else {
			auto expr2 = new ExprComponent(ExprComponentKind::binary, *token);
			expr2->op = token->value;
			expr2->S[0] = expr1;
			expr2->S[1] = expr_x;
			expr1 = expr2;
		}
	}
}

ExprComponent* CompileNotValidatedExpr::performFormula() {
	auto expr1 = performPrefix();
	if (expr1 == nullptr) {
		return nullptr;
	}
	return performBinary(expr1, 0);
}

ExprComponent* CompileNotValidatedExpr::performChunk(ExprComponent* expr1) {
	std::optional<Token> token;
	ExprComponent* ec = nullptr;
	if (expr1 == nullptr) {
		token = tokenizer->copyOrEnd();
		ec = new ExprComponent(ExprComponentKind::chunk, *token);
		ec->sp = new std::vector<ExprComponent*>;
	}
	else {
		ec = new ExprComponent(ExprComponentKind::chunk, expr1->token);
		ec->sp = new std::vector<ExprComponent*>;
		ec->sp->push_back(expr1);
	}
	while (1) {
		token = tokenizer->copyOrEnd();
		if (token.has_value()
			&& (token->kind != TokenKind::keyword)
			&& !token->isCloseBracket()
			&& !token->testRandom(U",")
			&& !token->testRandom(U";")
			&& !token->isBracket(U"{")) {
			expr1 = performFormula();
			if (expr1 == nullptr) {
				goto error;
			}
			ec->sp->push_back(expr1);
		}
		else {
			break;
		}
	}
	return ec;
error:
	delete ec;
	return nullptr;
}

ExprComponent* CompileNotValidatedExpr::performExpr() {
	auto expr1 = performFormula();
	if (expr1 == nullptr) {
		return nullptr;
	}
	auto token = tokenizer->copyOrEnd();
	if (token.has_value()
		&& (token->kind != TokenKind::keyword)
		&& !token->isCloseBracket()
		&& !token->testRandom(U",")
		&& !token->testRandom(U":")
		&& !token->testRandom(U";")
		&& !token->isBracket(U"{")) {
		return performChunk(expr1);
	}
	return expr1;
}

ExprComponent* CompileNotValidatedExpr::performBracket(const std::u32string& left, const std::u32string& right) {
	auto token = tokenizer->copyOrEnd();
	if (!token.has_value() || !token->isBracket(left)) {
		pushError(u8"未找到括号" + To_UTF8(left));
		return nullptr;
	}
	tokenizer->dropOrEnd();
	auto expr = new ExprComponent(ExprComponentKind::bracket, *token);
	std::u32string::value_type op_pair[2] = { pairBracket(token->value[0]) , 0 };
	expr->op = token->value + op_pair;
	expr->sp = performCommaChunk();
	if (expr->sp == nullptr) {
		delete expr;
		return nullptr;
	}
	token = tokenizer->copyOrEnd();
	if (!token.has_value() || !token->isBracket(right)) {
		pushError(u8"未找到括号" + To_UTF8(right));
		delete expr;
		return nullptr;
	}
	tokenizer->dropOrEnd();
	return expr;
}
	
std::vector<ExprComponent*>* CompileNotValidatedExpr::performCommaChunk() {
	auto sp = new std::vector<ExprComponent*>;
	std::optional<Token> token;
	while (1) {
		token = tokenizer->copyOrEnd();
		if (!token.has_value() 
			|| token->isCloseBracket() 
			|| token->testRandom(U";")
			|| token->isBracket(U"{")) {
			break;
		}
		if (sp->size() > 0) {
			if (!token->testRandom(U",")) {
				pushError(u8"未找到逗号");
				goto error;
			}
			token = tokenizer->dropAndCopy();
		}
		auto expr1 = performExpr();
		if (expr1 == nullptr) {
			goto error;
		}
		sp->push_back(expr1);
	}
	return sp;
error:
	if (sp != nullptr) {
		if (sp->size()) {
			for (auto s : *sp) {
				delete s;
			}
		}
		delete sp;
	}
	return nullptr;
}