#include "compiler.h"

#include "Modules/LLP.h"

#include "CompileNotValidatedImpl.h"

#include "Components/AutoPtr/AutoPtr.h"
#include "Components/Parser/Tokenizer.h"
#include "Components/PathConfig/PathConfig.h"
#include "Components/ModuleComponent/ExprComponent.h"
#include "Components/ModuleComponent/ModuleComponent.h"
#include "Components/ModuleCache/ModuleCache.h"

#include <list>

std::vector<std::u32string> CompileNotValidatedImpl::exprToPackage(ExprComponent* expr) {
	std::list<std::u32string> s;
	if ((expr == nullptr)
		|| (expr->sp == nullptr)
		|| (expr->sp->size() != 1)) {
		pushError(u8"不合法的包名称");
		goto error;
	}
	expr = (*expr->sp)[0];
	while (1) {
		if (expr->kind == ExprComponentKind::identifier) {
			s.push_front(expr->identifier);
			break;
		}
		else if ((expr->kind == ExprComponentKind::binary)
			&& (expr->op == U".")
			&& (expr->S[1] != nullptr)
			&& (expr->S[1]->kind == ExprComponentKind::identifier)){
			s.push_front(expr->S[1]->identifier);
			expr = expr->S[0];
		}
		else {
			pushError(u8"不合法的包名称");
			goto error;
		}
	}
	return std::vector<std::u32string>(s.begin(), s.end());
error:
	return {};
}

bool CompileNotValidatedImpl::performPackage(CompileProcess& compileProcess, ModuleComponent& mod) {
	std::vector<std::u32string> pt;
	auto token = tokenizer->copyOrEnd();
	if (!token.has_value() || !token->isKeyword(U"包")) {
		pushError(u8"程序内部调用错误");
		return false;
	}
	if (mod.package_name.size()) {
		pushError(u8"重复的包名称语句");
		return false;
	}
	tokenizer->dropOrEnd();
	auto ec = performBracket(U"(", U")");
	if (ec == nullptr) {
		pushError(u8"不合法的包名称");
		goto error;
	}
	pt = exprToPackage(ec);
	if (pt.size() == 0) {
		pushError(u8"不合法的包名称");
		goto error;
	}
	if (!PathConfig::checkPackage(pt, mod.file_path)) {
		pushError(u8"包名称错误");
		goto error;
	}
	token = tokenizer->copyOrEnd();
	if (!token.has_value() || !token->testRandom(U";")) {
		pushError(u8"未找到分号");
		goto error;
	}
	tokenizer->dropOrEnd();
	if (ec != nullptr) {
		delete ec;
	}
	{
		auto root = PathConfig::ResolvePackageRootPath(pt, mod.file_path);
		compileProcess.pathConfig.AddDir(root);
	}
	mod.package_name = pt;
	return true;
error:
	if (ec != nullptr) {
		delete ec;
	}
	return false;
}

bool CompileNotValidatedImpl::performImport(CompileProcess& compileProcess, ModuleComponent& mod) {
	std::vector<std::u32string> pt;
	auto token = tokenizer->copyOrEnd();
	if (!token.has_value() || !token->isKeyword(U"导入")) {
		pushError(u8"程序内部调用错误");
		return false;
	}
	tokenizer->dropOrEnd();
	auto ec = performBracket(U"(", U")");
	if (ec == nullptr) {
		pushError(u8"不合法的包名称");
		goto error;
	}
	pt = exprToPackage(ec);
	if (pt.size() == 0) {
		pushError(u8"不合法的包名称");
		goto error;
	}
	
	{
		auto file_import = compileProcess.pathConfig.searchPackage(pt);
		if (file_import.size() == 0) {
			pushError(u8"找不到需要导入的包");
			goto error;
		}
		if (queryLLP()->ImportModule(&mod, file_import) == false) {
			goto error;
		}
	}

	token = tokenizer->copyOrEnd();
	if (!token.has_value() || !token->testRandom(U";")) {
		pushError(u8"未找到分号");
		goto error;
	}
	tokenizer->dropOrEnd();
	return true;

error:
	if (ec != nullptr) {
		delete ec;
	}
	return false;

#if 0
	auto dot_expr = PackageGrammar::takeAnchor(tokenizer);
	if (!dot_expr.has_value() || dot_expr->size() == 0) {
		elp.pushError(tokenizer, u8"无效的导入说明符");
		return false;
	}
	auto file_path = elp.queryModuleLocator().searchModuleSource(*dot_expr);
	if (!file_path.has_value()) {
		elp.pushError(tokenizer, u8"导入的文件不存在");
		return false;
	}

	auto file_path_outer = tokenizer.queryScanner().getFileName();

	elpModule* outter_module = elp.queryModuleLocator().resolveFilePathToModule(file_path_outer);
	elpModule* import_module = elp.queryModuleLocator().resolveFilePathToModule(*file_path);
	if (import_module != nullptr) {
		if (testInImport(outter_module, import_module)) {
			elp.pushError(tokenizer, u8"无法解决的包导入");
			return false;
		}
		outter_module->addImportModule(import_module, true);
		return true;
	}

	import_module = new elpModule;
	elp.queryModuleLocator().setFileModule(*file_path, import_module);
	outter_module->addImportModule(import_module, true);
	return elp.queryCompiler().queryLLP()->importLLP(*file_path);
#endif
	return false;
}

bool CompileNotValidatedImpl::performClassDefine(CompileProcess& compileProcess, ModuleComponent& mod) {
	auto class_def = performClassDefineSyntax();
	if (class_def == nullptr) {
		return false;
	}
	mod.class_def.push_back(class_def);
	return true;
}

ModuleComponent* CompileNotValidatedImpl::performModule(CompileProcess& compileProcess) {
	
	ModuleComponent* mod = nullptr;

	if ((mod = queryLLP()->Runtime()->mcache->queryModuleCache(compileProcess.filePath)) != nullptr) {
		return mod;
	}

	SetCompileProcess(compileProcess);

	mod = new ModuleComponent;

	mod->file_path = compileProcess.filePath;

	queryLLP()->Runtime()->mcache->updateModuleCache(compileProcess.filePath, mod);

	auto token = tokenizer->copyOrEnd();

	if (token.has_value() && token->isKeyword(U"包")) {
		if (!this->performPackage(compileProcess, *mod)) {
			goto error;
		}
	}

	while (1) {
		token = tokenizer->copyOrEnd();
		if (!token.has_value() || !token->isKeyword(U"导入")) {
			break;
		}
		if (!this->performImport(compileProcess, *mod)) {
			goto error;
		}
	}

	while (1) {
		token = tokenizer->copyOrEnd();
		if (!token.has_value()) {
			break;
		}
		if (!this->performClassDefine(compileProcess, *mod)) {
			goto error;
		}
	}

	tokenizer->copyOrEnd();
	if (tokenizer->isEnd()) {
		return mod;
	}
	else {
		pushError(u8"非预期的输入");
		goto error;
	}
error:
	if (mod != nullptr) {
		delete mod;
	}
	return nullptr;
}

ExprComponent* CompileNotValidatedImpl::CompileNotValidedExprOnce(CompileProcess& compileProcess) {
	SetCompileProcess(compileProcess);
	return this->performExpr();
}

StatementComponent* CompileNotValidatedImpl::CompileNotValidedStatementOnce(CompileProcess& compileProcess) {
	SetCompileProcess(compileProcess);
	return this->performStatement();
}

FunctionDefinitionComponent* CompileNotValidatedImpl::CompileNotValidedFunctionOnce(CompileProcess& compileProcess) {
	SetCompileProcess(compileProcess);
	return this->performFunctionDefinition();
}

ClassDefinitionComponent* CompileNotValidatedImpl::CompileNotValidedClassOnce(CompileProcess& compileProcess) {
	SetCompileProcess(compileProcess);
	return this->performClassDefineSyntax();
}

ModuleComponent* CompileNotValidatedImpl::CompileNotValidatedModule(CompileProcess& compileProcess) {
	return this->performModule(compileProcess);
}