#pragma once

#include "Components/CompileProcess/CompileProcess.h"
#include "Components/ErrorBuffer/ErrorBuffer.h"
#include "Components/Parser/Tokenizer.h"

class ExprComponent;
class StatementComponent;
class FunctionDefinitionComponent;
class AttributeDefine;
class ClassDefinitionComponent;
class ModuleComponent;

class CompileNotValidatedImplBase
{
public:
	inline void SetCompileProcess(CompileProcess& cp) {
		compileProcess = &cp;
		if (compileProcess->tokenizer != nullptr) {
			setTokenizer(compileProcess->tokenizer);
		}
	}
	inline void setTokenizer(Tokenizer* tokenizer) {
		this->tokenizer = tokenizer;
	}
	inline void setErrorBuffer(ErrorBuffer* errorBuffer) {
		this->errorBuffer = errorBuffer;
	}
	inline void pushError(const std::u8string& s) {
		if (errorBuffer) {
			auto scanner = tokenizer->queryScanner();
			errorBuffer->push(scanner.getFileName(), scanner.getLine(), scanner.getColumn(), s);
		}
		else {
			auto scanner = tokenizer->queryScanner();
			SErrorBuffer.push(scanner.getFileName(), scanner.getLine(), scanner.getColumn(), s);
		}
	}
	CompileProcess* compileProcess = nullptr;
	Tokenizer* tokenizer = nullptr;
	ErrorBuffer* errorBuffer = nullptr;
};

class CompileNotValidatedExpr : public virtual CompileNotValidatedImplBase
{
public:
	ExprComponent* performPrefix();
	ExprComponent* performPostfix(ExprComponent* expr1);
	ExprComponent* performPrimary();
	ExprComponent* performBinary(ExprComponent* expr1, int precedence);
	ExprComponent* performFormula();
	ExprComponent* performChunk(ExprComponent* expr1);
public:
	ExprComponent* performExpr();
	ExprComponent* performBracket(const std::u32string& left, const std::u32string& right);
	std::vector<ExprComponent*>* performCommaChunk();
};

class CompileNotValidatedStatement : public virtual CompileNotValidatedImplBase, public virtual CompileNotValidatedExpr
{
public:
	StatementComponent* performExprStatment();
	StatementComponent* performIf();
	StatementComponent* performFor();
	StatementComponent* performWhile();
	StatementComponent* performJump();
	StatementComponent* performBlock();
	StatementComponent* performStatement();
};

class CompileNotValidatedFunctionDefine : public virtual CompileNotValidatedImplBase, public virtual CompileNotValidatedExpr, public virtual CompileNotValidatedStatement
{
public:
	FunctionDefinitionComponent* performFunctionDefinition(bool is_export, bool is_static, bool is_edge, ExprComponent* pt, std::u32string name);
	FunctionDefinitionComponent* performFunctionDefinition();
};

class CompileNotValidatedClassDefine : public virtual CompileNotValidatedImplBase, public virtual CompileNotValidatedExpr, public virtual CompileNotValidatedStatement, public virtual CompileNotValidatedFunctionDefine
{
public:
	std::vector<AttributeDefine*>* performClassAttributes();
	ClassDefinitionComponent* performClassDefineSyntax();
};

class CompileNotValidatedImpl : public virtual CompileNotValidatedExpr, public virtual CompileNotValidatedStatement, public virtual CompileNotValidatedFunctionDefine, public virtual CompileNotValidatedClassDefine
{
public:
	std::vector<std::u32string> exprToPackage(ExprComponent* expr);
public:
	bool performPackage(CompileProcess& compileProcess, ModuleComponent& mod);
	bool performImport(CompileProcess& compileProcess, ModuleComponent& mod);
	bool performClassDefine(CompileProcess& compileProcess, ModuleComponent& mod);
	ModuleComponent* performModule(CompileProcess& compileProcess);
public:
	ExprComponent* CompileNotValidedExprOnce(CompileProcess& compileProcess);
	StatementComponent* CompileNotValidedStatementOnce(CompileProcess& compileProcess);
	FunctionDefinitionComponent* CompileNotValidedFunctionOnce(CompileProcess& compileProcess);
	ClassDefinitionComponent* CompileNotValidedClassOnce(CompileProcess& compileProcess);
public:
	ModuleComponent* CompileNotValidatedModule(CompileProcess& compileProcess);
};