#include "compiler.h"

#include "ModuleValidation.h"

#include "Components/ModuleComponent/ClassDefinitionComponent.h"
#include "Components/ModuleComponent/ModuleComponent.h"
#include "Components/ModuleCache/ModuleCache.h"
#include "Components/NameBase/NameBase.h"

#include "Modules/LLP.h"

bool ModuleValidation::checkSpecifiedClassNotionSyntax(const std::u8string& file_name, ExprComponent* expr) {
	if (expr->kind == ExprComponentKind::identifier) {
		return true;
	}
	else if (expr->kind == ExprComponentKind::binary) {
		if (expr->S[0]->kind != ExprComponentKind::identifier) {
			pushError(file_name, expr->token, u8"不是一个可用的标识符");
			goto error;
		}
		if (expr->op != U"postfix.[]") {
			pushError(file_name, expr->token, u8"不是可用的类型记号");
			goto error;
		}
		if (expr->S[1]->kind != ExprComponentKind::bracket) {
			pushError(file_name, expr->token, u8"不是可用的类型记号");
			goto error;
		}
		for (auto s : *expr->S[1]->sp) {
			if (!checkSpecifiedClassNotionSyntax(file_name, s)) {
				goto error;
			}
		}
	}
	return true;
error:
	return false;
}

bool ModuleValidation::checkFunctionParamsListSyntax(const std::u8string& file_name, ExprComponent* expr) {
	if ((expr->kind != ExprComponentKind::bracket)
		|| (expr->sp == nullptr)) {
		pushError(file_name, expr->token, u8"不是一个可用的函数参数类别");
		goto error;
	}
	for (auto c : *expr->sp) {
		if (c->sp->size() != 2) {
			pushError(file_name, c->token, u8"不是一个可用的函数参数类别");
			goto error;
		}
		if (!checkSpecifiedClassNotionSyntax(file_name, (*c->sp)[0])) {
			pushError(file_name, c->token, u8"不是一个可用的函数参数类别");
			goto error;
		}
		if ((*c->sp)[1]->kind != ExprComponentKind::identifier) {
			pushError(file_name, c->token, u8"不是一个可用的函数参数类别");
			goto error;
		}
	}
	return true;
error:
	return false;
}

bool ModuleValidation::checkTypeCaptureSyntax(const std::u8string& file_name, ExprComponent* expr) {
	if (expr == nullptr) {
		SErrorBuffer.push(file_name, u8"程序内部错误");
		return false;
	}
	if ((expr->kind != ExprComponentKind::bracket)
		|| (expr->op != U"[]")
		|| (expr->sp == nullptr)) {
		pushError(file_name, expr->token, u8"不是一个可用的类型捕获式");
		goto error;
	}
	for (auto s : *(expr->sp)) {
		if (s->kind != ExprComponentKind::unary
			|| s->op != U"prefix.::") {
			pushError(file_name, s->token, u8"不是一个可用的类型捕获式");
			goto error;
		}
		if ((s->S == nullptr)
			|| s->S[0]->kind != ExprComponentKind::identifier) {
			pushError(file_name, s->token, u8"不是一个可用的类型捕获式");
			goto error;
		}
	}
	return true;
error:
	return false;
}

bool ModuleValidation::checkExpressionClassSyntax(const std::u8string& file_name, std::vector<ExprComponent*>& expr) {
	for (auto s : expr) {
		if (!checkSpecifiedClassNotionSyntax(file_name, s)) {
			pushError(file_name, s->token, u8"不是一个可用的类型记号");
			goto error;
		}
	}
	return true;
error:
	return false;
}

bool ModuleValidation::checkImportModuleNotNesting(ModuleComponent* new_import, ModuleComponent* outer) {
	if (outer != nullptr) {
		for (auto s : new_import->import_module) {
			auto testMod = queryLLP()->Runtime()->mcache->queryModuleCache(s);
			if (outer == testMod) {
				return false;
			}
			if (!checkImportModuleNotNesting(testMod, outer)) {
				return false;
			}
		}
	}
	return true;
}

bool ModuleValidation::CheckClassPremiseEq(ClassDefinitionComponent* a, ClassDefinitionComponent* b) {
	if (a->name != b->name) {
		return false;
	}
	if ((a->is_export != b->is_export)
		|| (a->is_second != b->is_second)
		|| (a->is_template != b->is_template)
		|| (a->is_completed != b->is_completed)) {
		return false;
	}
	return true;
}

bool ModuleValidation::LoadModuleSymbol(NameBase& nb, ModuleComponent& Module) {
	nb.putScope(ScopeKind::module_def, &Module, nullptr, nullptr);
	for (auto cdef : Module.class_def) {
		auto fc = nb.getEntry(cdef->name);
		if (cdef->attrs != nullptr) {
			if (fc != nullptr) {
				if ((fc->kind != ScopeEntryKind::class_entry)||(fc->c->attrs != nullptr)) {
					pushError(Module.file_path, cdef->token, u8"类定义冲突");
					return false;
				}
			}
		}
		else if(fc != nullptr){
			if (fc->kind != ScopeEntryKind::class_entry) {
				pushError(Module.file_path, cdef->token, u8"类定义冲突");
				return false;
			}
			if (!CheckClassPremiseEq(fc->c, cdef)){
				pushError(Module.file_path, cdef->token, u8"类定义冲突");
				return false;
			}
			continue;
		}
		nb.putEntry(cdef->name, cdef);
	}
	return true;
}

bool ModuleValidation::ValidateClassDefinition(NameBase& nb, ModuleComponent* mod, ClassDefinitionComponent* cdef) {
	return true;
}

ModuleComponent* ModuleValidation::ValidateModule(ModuleComponent & NotValidatedModule){
	if (NotValidatedModule.validated) {
		return &NotValidatedModule;
	}
	std::vector<ModuleComponent*> import_mod;
	for (auto s : NotValidatedModule.import_module) {
		auto im = queryLLP()->Runtime()->mcache->queryModuleCache(s);
		if (im->validated == false) {
			if (ValidateModule(*im) == nullptr) {
				return nullptr;
			}
		}
		import_mod.push_back(im);
	}
	NameBase nb;
	for (auto s : import_mod) {
		if (!LoadModuleSymbol(nb, *s)) {
			return nullptr;
		}
	}
	if (!LoadModuleSymbol(nb, NotValidatedModule)) {
		return nullptr;
	}
	for (auto cdef : NotValidatedModule.class_def) {
		if (!ValidateClassDefinition(nb, &NotValidatedModule, cdef)) {
			return nullptr;
		}
	}
	NotValidatedModule.validated = true;
	return &NotValidatedModule;
}