#include "Tokenizer.h"

#include "Components/DefaultConfig/DefaultConfig.h"
#include "Components/ErrorBuffer/ErrorBuffer.h"
#include "Components/PathConfig/PathConfig.h"

Tokenizer::~Tokenizer() {
	if (scanner != nullptr) {
		delete scanner;
	}
}

Scanner& Tokenizer::queryScanner() {
	return *scanner;
}


Tokenizer::Tokenizer() {

}

void Tokenizer::initTokenizer(const std::u8string& file_path, const std::u32string& chunk, ErrorBuffer* error_buffer) {
	if (scanner == nullptr) {
		scanner = new Scanner;
	}
	scanner->initScanner(file_path, chunk);
	this->error_buffer = error_buffer;
	dropOrEnd();
}

std::optional<Token> Tokenizer::copyOrEnd() {
	return token;
}

std::optional<Token> Tokenizer::dropAndCopy() {
	dropOrEnd();
	return copyOrEnd();
}

bool skip_block_comment(Scanner& scanner) {
	if (scanner.testSequence(U"/*")) {
		scanner.dropOrEnd(); scanner.dropOrEnd();
		scanner.skipUntilStringOrEnd(U"*/");
		if (scanner.testSequence(U"*/")) {
			scanner.dropOrEnd(); scanner.dropOrEnd();
			return true;
		}
		else {
			return false;
		}
	}
	return false;
}

bool skip_indent(Scanner& scanner, int n) {
	for (int i = 0; i < n; ++i) {
		int c = scanner.copyOrEnd();
		if ((c != ' ') && (c != '\t')) {
			return false;
		}
		scanner.dropOrEnd();
	}
	return true;
}

bool skip_continue_comment_or_end(Scanner& scanner) {
	if (scanner.testSequence(U"//")) {
		int count_line = 1;
		int ctrl, line_pos, column_pos;
		line_pos = scanner.getLine();
		column_pos = scanner.getColumn();
		ctrl = scanner.getIndent();
		if (ctrl + 1 != column_pos) {
			goto error_invalid_continue_comment;
		}
		scanner.skipLineOrEnd();
		while (1) {
			if (!scanner.testSequence(U"//")) {
				if (count_line >= 3) {
					return true;
				}
				goto error_invalid_continue_comment;
			}
			scanner.skipLineOrEnd();
			++count_line;
		}
	}
	return false;
error_invalid_continue_comment:
	return false;
}

void Tokenizer::dropOrEnd() {
	if (scanner->isEnd()) {
		if (token.has_value()) {
			token.reset();
		}
		return;
	}
	while (1) {
		char32_t c = scanner->copyOrEnd();
		if (c == 0) {
			if (token.has_value()) {
				token.reset();
			}
			return;
		}
		if ((c == ' ')
			|| (c == '\t')
			|| (c == '\r')
			|| (c == '\n')) {
			scanner->dropOrEnd();
			continue;
		}
		if (scanner->testSequence(U"//")) {
			if (!skip_continue_comment_or_end(*scanner)) {
				int line_pos, column_pos;
				line_pos = scanner->getLine();
				column_pos = scanner->getColumn();
				pushError(line_pos, column_pos, u8"非法的连续注释");
				scanner->setEnd();
				goto next_token_error;
			}
			continue;
		}
		if (scanner->testSequence(U"/*")) {
			if (!skip_block_comment(*scanner)) {
				int line_pos, column_pos;
				line_pos = scanner->getLine();
				column_pos = scanner->getColumn();
				pushError(line_pos, column_pos, u8"非法的块注释");
				scanner->setEnd();
				goto next_token_error;
			}
			continue;
		}
		for(int i=0; i<2; ++i){
			auto kind = SLexicalRule.BreakingSet[i].kind;
			for (auto& s : SLexicalRule.BreakingSet[i].patterns) {
				if (scanner->testSequence(s)) {
					int line_pos, column_pos;
					line_pos = scanner->getLine();
					column_pos = scanner->getColumn();
					this->token = Token(
						kind,
						s,
						line_pos,
						column_pos
					);
					scanner->skipCountOrEnd(s.size());
					return;
				}
			}
		}
		{
			std::u32string sp;
			char32_t c = scanner->copyOrEnd();
			if (SLexicalRule.isIdentifierLetter(c)) {
				int line_pos, column_pos;
				line_pos = scanner->getLine();
				column_pos = scanner->getColumn();
				while (1) {
					char32_t c = scanner->copyOrEnd();
					if (SLexicalRule.isIdentifierLetter(c)) {
						sp.push_back(c);
						scanner->dropOrEnd();
						continue;
					}
					else {
						break;
					}
				}
				for (auto& s : SLexicalRule.Keywords) {
					if (sp == s) {
						int line_pos, column_pos;
						line_pos = scanner->getLine();
						column_pos = scanner->getColumn();
						this->token = Token(
							TokenKind::keyword,
							s,
							line_pos,
							column_pos
						);
						return;
					}
				}
				this->token = Token(
					TokenKind::identifier,
					sp,
					line_pos,
					column_pos
				);
				return;
			}
		}
		int line_pos, column_pos;
		line_pos = scanner->getLine();
		column_pos = scanner->getColumn();
		scanner->setEnd();
		pushError(line_pos, column_pos, u8"无法识别的输入");
	next_token_error:
		if (token.has_value()) {
			token.reset();
		}
		return;
	}
}

bool Tokenizer::isEnd() {
	return !token.has_value();
}

void Tokenizer::setEnd() {
	scanner->setEnd();
	dropOrEnd();
}

void Tokenizer::pushError(int line_pos, int column_pos, const std::u8string& error_string) {
	error_buffer->push(scanner->getFileName(), line_pos, column_pos, error_string);
}

Tokenizer* Tokenizer::createTokenizer(const std::u8string& file_path, ErrorBuffer* error_buffer) {
	std::u32string content = PathConfig::loadFileContent(file_path);
	Tokenizer* tokenizer = new Tokenizer;
	tokenizer->initTokenizer(file_path, content, error_buffer);
	return tokenizer;
}