#include "compiler.h"

#include "LLP.h"
#include "ELP.h"

#include "Components/ModuleCache/ModuleCache.h"
#include "Components/ModuleValidation/ModuleValidation.h"
#include "Components/ModuleComponent/ModuleComponent.h"
#include "Components/CppBuilder/CppBuilder.h"

#include "Components/AutoPtr/AutoPtr.h"
#include "Components/CompileProcess/CompileProcess.h"
#include "Components/Parser/Tokenizer.h"
#include "Components/CompileNotValidated/CompileNotValidated.h"
#include "Components/ModuleValidation/ModuleValidation.h"
#include "Components/Inspector/Inspector.h"

#include "unicode.h"

#include <iostream>
#include <fstream>

/*===-- LogicLanguageProcessEc --===*/

LogicLanguageProcessRuntime::LogicLanguageProcessRuntime(){
	mcache = new ModuleCache;
	valid = new ModuleValidation;
}

LogicLanguageProcessRuntime::~LogicLanguageProcessRuntime() {
	if (mcache != nullptr) { delete mcache;  }
}

/*===-- LogicLanguageProcess --===*/

LogicLanguageProcessRuntime* LogicLanguageProcess::Runtime_ = nullptr;

LogicLanguageProcessRuntime* LogicLanguageProcess::Runtime() {
	if (Runtime_ == nullptr) {
		Runtime_ = new LogicLanguageProcessRuntime;
	}
	return Runtime_;
}

void LogicLanguageProcess::printExprOnce(const std::u8string& file_path) {
	AutoPtr<CompileProcess> compileProcess(new CompileProcess(file_path, PathConfig()));
	compileProcess.data().tokenizer = Tokenizer::createTokenizer(file_path, &SErrorBuffer);
	CompileNotValidated compile1;
	auto expr = compile1.CompileNotValidedExprOnce(compileProcess.data());
	if (expr == nullptr) {
		goto error;
	}
	Inspector inspector;
	inspector.print(expr);
error:
	SErrorBuffer.flush();
}

void LogicLanguageProcess::printStatementOnce(const std::u8string& file_path) {
	AutoPtr<CompileProcess> compileProcess(new CompileProcess(file_path, PathConfig()));
	compileProcess.data().tokenizer = Tokenizer::createTokenizer(file_path, &SErrorBuffer);
	CompileNotValidated compile1;
	auto expr = compile1.CompileNotValidedStatementOnce(compileProcess.data());
	if (expr == nullptr) {
		goto error;
	}
	Inspector inspector;
	inspector.print(expr);
error:
	SErrorBuffer.flush();

}

void LogicLanguageProcess::printFunctionDefineOnce(const std::u8string& file_path) {
	AutoPtr<CompileProcess> compileProcess(new CompileProcess(file_path, PathConfig()));
	compileProcess.data().tokenizer = Tokenizer::createTokenizer(file_path, &SErrorBuffer);
	CompileNotValidated compile1;
	auto expr = compile1.CompileNotValidedFunctionOnce(compileProcess.data());
	if (expr == nullptr) {
		goto error;
	}
	Inspector inspector;
	inspector.print(expr);
error:
	SErrorBuffer.flush();
}

void LogicLanguageProcess::printClassDefineOnce(const std::u8string& file_path) {
	AutoPtr<CompileProcess> compileProcess(new CompileProcess(file_path, PathConfig()));
	compileProcess.data().tokenizer = Tokenizer::createTokenizer(file_path, &SErrorBuffer);
	CompileNotValidated compile1;
	auto expr = compile1.CompileNotValidedClassOnce(compileProcess.data());
	if (expr == nullptr) {
		goto error;
	}
	Inspector inspector;
	inspector.print(expr);
error:
	SErrorBuffer.flush();
}

bool LogicLanguageProcess::ImportModule(ModuleComponent* outerModule, const std::u8string& file_path) {
	outerModule->import_module.insert(file_path);

	auto mod = Runtime()->mcache->queryModuleCache(file_path);

	CompileProcess* compileProcess = nullptr;

	if (mod != nullptr) {
		if(!queryLLP()->Runtime()->valid->checkImportModuleNotNesting(mod, outerModule)){
			SErrorBuffer.push(outerModule->file_path, u8"无限嵌套的模块导入 <" + outerModule->file_path + u8"> 和 <" + mod->file_path + u8">");
			goto error;
		}
		return true;
	}

	compileProcess = new CompileProcess(file_path, PathConfig());

	compileProcess->tokenizer = Tokenizer::createTokenizer(file_path, &SErrorBuffer);
	CompileNotValidated compile1;
	mod = compile1.CompileNotValidatedModule(*compileProcess);
	if (mod == nullptr) {
		goto error;
	}
	mod = queryLLP()->Runtime()->valid->ValidateModule(*mod);
	if (mod == nullptr) {
		goto error;
	}
	Runtime()->mcache->updateModuleCache(file_path, mod);

	if (!queryLLP()->Runtime()->valid->checkImportModuleNotNesting(mod, outerModule)) {
		SErrorBuffer.push(outerModule->file_path, u8"无限嵌套的模块导入 <" + outerModule->file_path + u8"> 和 <" + mod->file_path + u8">");
		goto error;
	}

	return true;
error:
	SErrorBuffer.flush();
	return false;
}

ModuleComponent* LogicLanguageProcess::CompileModule(const std::u8string& file_path) {
	auto mod = Runtime()->mcache->queryModuleCache(file_path);
	if (mod != nullptr) {
		return mod;
	}

	AutoPtr<CompileProcess> compileProcess(new CompileProcess(file_path, PathConfig()));

	compileProcess.data().tokenizer = Tokenizer::createTokenizer(file_path, &SErrorBuffer);
	CompileNotValidated compile1;
	mod = compile1.CompileNotValidatedModule(compileProcess.data());
	if (mod == nullptr) {
		goto error;
	}
	mod = queryLLP()->Runtime()->valid->ValidateModule(*mod);
	if (mod == nullptr) {
		goto error;
	}
	Runtime()->mcache->updateModuleCache(file_path, mod);
	return mod;
error:
	SErrorBuffer.flush();
	return nullptr;
}

bool searchArgvOption(int argc, char** argv, const char* name) {
	for (int i = 0; i < argc; ++i) {
		if (std::string(argv[i]) == std::string(name)) {
			return true;
		}
	}
	return false;
}

std::u8string searchArgv(int argc, char** argv, const char* name) {
	for (int i = 0; i < argc-1; ++i) {
		if (std::string(argv[i]) == std::string(name)) {
			return To_UTF8(std::string(argv[i + 1]));
		}
	}
	return {};
}

int LogicLanguageProcess::Build(int argc, char** argv) {
	std::u8string file_compile = searchArgv(argc, argv, "-compile");
	auto mod = CompileModule(file_compile);
	if (mod != nullptr) {
		CppBuilder builder;
		{
			std::u8string chead = searchArgv(argc, argv, "-genCppHeadFile");
			if (chead.size()) {
				std::fstream fout(To_CppString(chead), std::ios_base::out);
				if (!fout.is_open()) {
					std::cerr << "文件 <" << To_CppString(chead) << "> 无法打开\n";
					return -1;
				}
				auto codeHeadFile = builder.buildHeadFile(mod, searchArgvOption(argc, argv, "-primitiveTypes"));
				fout << To_CppString(codeHeadFile);
				fout.close();
			}
		}
		{
			std::u8string cpp = searchArgv(argc, argv, "-genCppSourceFile");
			if (cpp.size()) {
				std::fstream fout(To_CppString(cpp), std::ios_base::out);
				if (!fout.is_open()) {
					std::cerr << "文件 <" << To_CppString(cpp) << "> 无法打开\n";
					return -1;
				}
				auto codeCpp = builder.buildCppFile(mod);
				fout << To_CppString(codeCpp);
				fout.close();
			}
		}
		//printClassDefineOnce(u8"C:\\Users\\Administrator\\Desktop\\ShuiPackage\\Source\\TestPackage\\test.shui");
	}
	return 0;
}